@@ -98,9 +98,7 @@ static void destroyPrivate()
 // Convert a string into a watch Method
 static KDirWatch::Method methodFromString(const QByteArray &method)
 {
-    if (method == "Fam") {
-        return KDirWatch::FAM;
-    } else if (method == "Stat") {
+    if (method == "Stat") {
         return KDirWatch::Stat;
     } else if (method == "QFSWatch") {
         return KDirWatch::QFSWatch;
@@ -117,8 +115,6 @@ static KDirWatch::Method methodFromString(const QByteArray &method)
 static const char *methodToString(KDirWatch::Method method)
 {
     switch (method) {
-    case KDirWatch::FAM:
-        return "Fam";
     case KDirWatch::INotify:
         return "INotify";
     case KDirWatch::Stat:
@@ -164,7 +160,7 @@ static const char s_envNfsMethod[] = "KDIRWATCH_NFSMETHOD";
  */
 
 KDirWatchPrivate::KDirWatchPrivate()
-    : timer()
+    : m_statRescanTimer()
     , freq(3600000)
     , // 1 hour as upper bound
     statEntries(0)
@@ -177,54 +173,42 @@ KDirWatchPrivate::KDirWatchPrivate()
     ,
 #endif
     _isStopped(false)
-    , m_references(0)
 {
     // Debug unittest on CI
     if (qAppName() == QLatin1String("kservicetest") || qAppName() == QLatin1String("filetypestest")) {
         s_verboseDebug = true;
     }
-    timer.setObjectName(QStringLiteral("KDirWatchPrivate::timer"));
-    connect(&timer, &QTimer::timeout, this, &KDirWatchPrivate::slotRescan);
+    m_statRescanTimer.setObjectName(QStringLiteral("KDirWatchPrivate::timer"));
+    connect(&m_statRescanTimer, &QTimer::timeout, this, &KDirWatchPrivate::slotRescan);
 
     m_nfsPollInterval = qEnvironmentVariableIsSet(s_envNfsPoll) ? qEnvironmentVariableIntValue(s_envNfsPoll) : 5000;
     m_PollInterval = qEnvironmentVariableIsSet(s_envPoll) ? qEnvironmentVariableIntValue(s_envPoll) : 500;
 
     m_preferredMethod = methodFromString(qEnvironmentVariableIsSet(s_envMethod) ? qgetenv(s_envMethod) : "inotify");
     // The nfs method defaults to the normal (local) method
-    m_nfsPreferredMethod = methodFromString(qEnvironmentVariableIsSet(s_envNfsMethod) ? qgetenv(s_envNfsMethod) : "Fam");
+    m_nfsPreferredMethod = methodFromString(qEnvironmentVariableIsSet(s_envNfsMethod) ? qgetenv(s_envNfsMethod) : "Stat");
 
     QList<QByteArray> availableMethods;
 
     availableMethods << "Stat";
 
-    // used for FAM and inotify
+    // used for inotify
     rescan_timer.setObjectName(QStringLiteral("KDirWatchPrivate::rescan_timer"));
     rescan_timer.setSingleShot(true);
     connect(&rescan_timer, &QTimer::timeout, this, &KDirWatchPrivate::slotRescan);
 
-#if HAVE_FAM
-    availableMethods << "FAM";
-    use_fam = true;
-    sn = nullptr;
-#endif
-
 #if HAVE_SYS_INOTIFY_H
-    supports_inotify = true;
-
     m_inotify_fd = inotify_init();
+    supports_inotify = m_inotify_fd > 0;
 
-    if (m_inotify_fd <= 0) {
+    if (!supports_inotify) {
         qCDebug(KDIRWATCH) << "Can't use Inotify, kernel doesn't support it:" << strerror(errno);
-        supports_inotify = false;
-    }
-
-    // qCDebug(KDIRWATCH) << "INotify available: " << supports_inotify;
-    if (supports_inotify) {
+    } else {
         availableMethods << "INotify";
         (void)fcntl(m_inotify_fd, F_SETFD, FD_CLOEXEC);
 
         mSn = new QSocketNotifier(m_inotify_fd, QSocketNotifier::Read, this);
-        connect(mSn, SIGNAL(activated(int)), this, SLOT(inotifyEventReceived()));
+        connect(mSn, &QSocketNotifier::activated, this, &KDirWatchPrivate::inotifyEventReceived);
     }
 #endif
 #if HAVE_QFILESYSTEMWATCHER
@@ -232,21 +216,27 @@ KDirWatchPrivate::KDirWatchPrivate()
     fsWatcher = nullptr;
 #endif
 
-    if (s_verboseDebug) {
-        qCDebug(KDIRWATCH) << "Available methods: " << availableMethods << "preferred=" << methodToString(m_preferredMethod);
-    }
+    qCDebug(KDIRWATCH) << "Available methods: " << availableMethods << "preferred=" << methodToString(m_preferredMethod);
 }
 
 // This is called on app exit (deleted by QThreadStorage)
 KDirWatchPrivate::~KDirWatchPrivate()
 {
-    timer.stop();
+    m_statRescanTimer.stop();
 
-#if HAVE_FAM
-    if (use_fam && sn) {
-        FAMClose(&fc);
+    // Unset us as d pointer. This indicates to the KDirWatch that the private has already been destroyed and therefore
+    // needs no additional cleanup from its destructor.
+    for (auto it = m_mapEntries.begin(); it != m_mapEntries.end(); it++) {
+        auto &entry = it.value();
+        for (auto &client : entry.m_clients) {
+            client.instance->d = nullptr;
+        }
     }
-#endif
+    // Also cover all instance that hold a reference to us (even when they don't have an entry)
+    for (auto &referenceObject : m_referencesObjects) {
+        referenceObject->d = nullptr;
+    }
+
 #if HAVE_SYS_INOTIFY_H
     if (supports_inotify) {
         QT_CLOSE(m_inotify_fd);
@@ -323,14 +313,10 @@ void KDirWatchPrivate::inotifyEventReceived()
 
             const QString tpath = e->path + QLatin1Char('/') + path;
 
-            if (s_verboseDebug) {
-                qCDebug(KDIRWATCH).nospace() << "got event 0x" << qPrintable(QString::number(event->mask, 16)) << " for " << e->path;
-            }
+            qCDebug(KDIRWATCH).nospace() << "got event " << inotifyEventName(event) << " for entry " << e->path
+                                         << (event->mask & IN_ISDIR ? " [directory] " : " [file] ") << path;
 
             if (event->mask & IN_DELETE_SELF) {
-                if (s_verboseDebug) {
-                    qCDebug(KDIRWATCH) << "-->got deleteself signal for" << e->path;
-                }
                 e->m_status = NonExistent;
                 m_inotify_wd_to_entry.remove(e->wd);
                 e->wd = -1;
@@ -351,12 +337,8 @@ void KDirWatchPrivate::inotifyEventReceived()
             if (event->mask & (IN_CREATE | IN_MOVED_TO)) {
                 Entry *sub_entry = e->findSubEntry(tpath);
 
-                if (s_verboseDebug) {
-                    qCDebug(KDIRWATCH) << "-->got CREATE signal for" << (tpath) << "sub_entry=" << sub_entry;
-                    qCDebug(KDIRWATCH) << *e;
-                }
+                qCDebug(KDIRWATCH) << "-->got CREATE signal for" << (tpath) << "sub_entry=" << sub_entry;
 
-                // The code below is very similar to the one in checkFAMEvent...
                 if (sub_entry) {
                     // We were waiting for this new file/dir to be created
                     sub_entry->dirty = true;
@@ -381,9 +363,6 @@ void KDirWatchPrivate::inotifyEventReceived()
                 }
             }
             if (event->mask & (IN_DELETE | IN_MOVED_FROM)) {
-                if (s_verboseDebug) {
-                    qCDebug(KDIRWATCH) << "-->got DELETE signal for" << tpath;
-                }
                 if ((e->isDir) && (!e->m_clients.empty())) {
                     // A file in this directory has been removed.  It wasn't an explicitly
                     // watched file as it would have its own watch descriptor, so
@@ -401,9 +380,6 @@ void KDirWatchPrivate::inotifyEventReceived()
             }
             if (event->mask & (IN_MODIFY | IN_ATTRIB)) {
                 if ((e->isDir) && (!e->m_clients.empty())) {
-                    if (s_verboseDebug) {
-                        qCDebug(KDIRWATCH) << "-->got MODIFY signal for" << (tpath);
-                    }
                     // A file in this directory has been changed.  No
                     // addEntry/ removeEntry bookkeeping should be required.
                     // Add the path to the list of pending file changes if
@@ -442,7 +418,7 @@ KDirWatchPrivate::Entry::~Entry()
 {
 }
 
-/* In FAM mode, only entries which are marked dirty are scanned.
+/* In inotify mode, only entries which are marked dirty are scanned.
  * We first need to mark all yet nonexistent, but possible created
  * entries as dirty...
  */
@@ -539,6 +515,60 @@ QList<const KDirWatchPrivate::Client *> KDirWatchPrivate::Entry::inotifyClientsF
     return ret;
 }
 
+QDebug operator<<(QDebug debug, const KDirWatch & /* watch */)
+{
+    if (!dwp_self.hasLocalData()) {
+        debug << "KDirWatch not used";
+        return debug;
+    }
+    debug << dwp_self.localData();
+    return debug;
+}
+
+QDebug operator<<(QDebug debug, const KDirWatchPrivate &dwp)
+{
+    debug << "Entries watched:";
+    if (dwp.m_mapEntries.count() == 0) {
+        debug << "  None.";
+    } else {
+        auto it = dwp.m_mapEntries.cbegin();
+        for (; it != dwp.m_mapEntries.cend(); ++it) {
+            const KDirWatchPrivate::Entry &e = it.value();
+            debug << "  " << e;
+
+            for (const KDirWatchPrivate::Client &c : e.m_clients) {
+                QByteArray pending;
+                if (c.watchingStopped) {
+                    if (c.pending & KDirWatchPrivate::Deleted) {
+                        pending += "deleted ";
+                    }
+                    if (c.pending & KDirWatchPrivate::Created) {
+                        pending += "created ";
+                    }
+                    if (c.pending & KDirWatchPrivate::Changed) {
+                        pending += "changed ";
+                    }
+                    if (!pending.isEmpty()) {
+                        pending = " (pending: " + pending + ')';
+                    }
+                    pending = ", stopped" + pending;
+                }
+                debug << "    by " << c.instance->objectName() << " (" << c.count << " times)" << pending;
+            }
+            if (!e.m_entries.isEmpty()) {
+                debug << "    dependent entries:";
+                for (KDirWatchPrivate::Entry *d : e.m_entries) {
+                    debug << "      " << d << d->path << (d->m_status == KDirWatchPrivate::NonExistent ? "NonExistent" : "EXISTS this is an ERROR!");
+                    if (s_verboseDebug) {
+                        Q_ASSERT(d->m_status == KDirWatchPrivate::NonExistent); // it doesn't belong here otherwise
+                    }
+                }
+            }
+        }
+    }
+    return debug;
+}
+
 QDebug operator<<(QDebug debug, const KDirWatchPrivate::Entry &entry)
 {
     debug.nospace() << "[ Entry for " << entry.path << ", " << (entry.isDir ? "dir" : "file");
@@ -546,8 +576,7 @@ QDebug operator<<(QDebug debug, const KDirWatchPrivate::Entry &entry)
         debug << ", non-existent";
     }
     debug << ", using "
-          << ((entry.m_mode == KDirWatchPrivate::FAMMode)            ? "FAM"
-                  : (entry.m_mode == KDirWatchPrivate::INotifyMode)  ? "INotify"
+          << ((entry.m_mode == KDirWatchPrivate::INotifyMode)        ? "INotify"
                   : (entry.m_mode == KDirWatchPrivate::QFSWatchMode) ? "QFSWatch"
                   : (entry.m_mode == KDirWatchPrivate::StatMode)     ? "Stat"
                                                                      : "Unknown Method");
@@ -570,8 +599,7 @@ QDebug operator<<(QDebug debug, const KDirWatchPrivate::Entry &entry)
 
 KDirWatchPrivate::Entry *KDirWatchPrivate::entry(const QString &_path)
 {
-    // we only support absolute paths
-    if (_path.isEmpty() || QDir::isRelativePath(_path)) {
+    if (_path.isEmpty()) {
         return nullptr;
     }
 
@@ -593,98 +621,13 @@ void KDirWatchPrivate::useFreq(Entry *e, int newFreq)
     // a reasonable frequency for the global polling timer
     if (e->freq < freq) {
         freq = e->freq;
-        if (timer.isActive()) {
-            timer.start(freq);
+        if (m_statRescanTimer.isActive()) {
+            m_statRescanTimer.start(freq);
         }
         qCDebug(KDIRWATCH) << "Global Poll Freq is now" << freq << "msec";
     }
 }
 
-#if HAVE_FAM
-// setup FAM notification, returns false if not possible
-bool KDirWatchPrivate::useFAM(Entry *e)
-{
-    if (!use_fam) {
-        return false;
-    }
-
-    if (!sn) {
-        if (FAMOpen(&fc) == 0) {
-            sn = new QSocketNotifier(FAMCONNECTION_GETFD(&fc), QSocketNotifier::Read, this);
-            connect(sn, SIGNAL(activated(int)), this, SLOT(famEventReceived()));
-        } else {
-            use_fam = false;
-            return false;
-        }
-    }
-
-    // handle FAM events to avoid deadlock
-    // (FAM sends back all files in a directory when monitoring)
-    famEventReceived();
-
-    e->m_mode = FAMMode;
-    e->dirty = false;
-    e->m_famReportedSeen = false;
-
-    bool startedFAMMonitor = false;
-
-    if (e->isDir) {
-        if (e->m_status == NonExistent) {
-            // If the directory does not exist we watch the parent directory
-            addEntry(nullptr, e->parentDirectory(), e, true);
-        } else {
-            int res = FAMMonitorDirectory(&fc, QFile::encodeName(e->path).data(), &(e->fr), e);
-            startedFAMMonitor = true;
-            if (res < 0) {
-                e->m_mode = UnknownMode;
-                use_fam = false;
-                delete sn;
-                sn = nullptr;
-                return false;
-            }
-            qCDebug(KDIRWATCH).nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr)) << ") for " << e->path;
-        }
-    } else {
-        if (e->m_status == NonExistent) {
-            // If the file does not exist we watch the directory
-            addEntry(nullptr, QFileInfo(e->path).absolutePath(), e, true);
-        } else {
-            int res = FAMMonitorFile(&fc, QFile::encodeName(e->path).data(), &(e->fr), e);
-            startedFAMMonitor = true;
-            if (res < 0) {
-                e->m_mode = UnknownMode;
-                use_fam = false;
-                delete sn;
-                sn = nullptr;
-                return false;
-            }
-
-            qCDebug(KDIRWATCH).nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr)) << ") for " << e->path;
-        }
-    }
-
-    // handle FAM events to avoid deadlock
-    // (FAM sends back all files in a directory when monitoring)
-    const int iterationCap = 80;
-    for (int i = 0; i <= iterationCap; ++i) { // we'll not wait forever; blocking for 4s seems plenty
-        famEventReceived();
-        // NB: check use_fam, if fam is defunct event receiving might disable fam support!
-        if (use_fam && startedFAMMonitor && !e->m_famReportedSeen) {
-            // 50 is ~half the time it takes to setup a watch.  If gamin's latency
-            // gets better, this can be reduced.
-            QThread::msleep(50);
-        } else if (use_fam && i == iterationCap) {
-            disableFAM();
-            return false;
-        } else {
-            break;
-        }
-    }
-
-    return true;
-}
-#endif
-
 #if HAVE_SYS_INOTIFY_H
 // setup INotify notification, returns false if not possible
 bool KDirWatchPrivate::useINotify(Entry *e)
@@ -763,7 +706,7 @@ bool KDirWatchPrivate::useStat(Entry *e)
 
         if (statEntries == 1) {
             // if this was first STAT entry (=timer was stopped)
-            timer.start(freq); // then start the timer
+            m_statRescanTimer.start(freq); // then start the timer
             qCDebug(KDIRWATCH) << " Started Polling Timer, freq " << freq;
         }
     }
@@ -890,6 +833,7 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path, Entry
     }
 
     if (exists && e->isDir && (watchModes != KDirWatch::WatchDirOnly)) {
+        // recursive watch for folders
         QFlags<QDir::Filter> filters = QDir::NoDotAndDotDot;
 
         if ((watchModes & KDirWatch::WatchSubDirs) && (watchModes & KDirWatch::WatchFiles)) {
@@ -901,7 +845,7 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path, Entry
         }
 
 #if HAVE_SYS_INOTIFY_H
-        if (e->m_mode == INotifyMode || (e->m_mode == UnknownMode && m_preferredMethod == KDirWatch::INotify)) {
+        if (m_preferredMethod == KDirWatch::INotify) {
             // qCDebug(KDIRWATCH) << "Ignoring WatchFiles directive - this is implicit with inotify";
             // Placing a watch on individual files is redundant with inotify
             // (inotify gives us WatchFiles functionality "for free") and indeed
@@ -935,9 +879,7 @@ void KDirWatchPrivate::addWatch(Entry *e)
     // are made locally. #177892.
 
     KDirWatch::Method preferredMethod = m_preferredMethod;
-    if (KNetworkMounts::self()->isOptionEnabledForPath(e->path, KNetworkMounts::KDirWatchUseINotify)) {
-        preferredMethod = KDirWatch::INotify;
-    } else if (m_nfsPreferredMethod != m_preferredMethod) {
+    if (m_nfsPreferredMethod != m_preferredMethod) {
         if (KFileSystemType::fileSystemType(e->path) == KFileSystemType::Nfs) {
             preferredMethod = m_nfsPreferredMethod;
         }
@@ -946,15 +888,6 @@ void KDirWatchPrivate::addWatch(Entry *e)
     // Try the appropriate preferred method from the config first
     bool entryAdded = false;
     switch (preferredMethod) {
-#if HAVE_FAM
-    case KDirWatch::FAM:
-        entryAdded = useFAM(e);
-        break;
-#else
-    case KDirWatch::FAM:
-        entryAdded = false;
-        break;
-#endif
 #if HAVE_SYS_INOTIFY_H
     case KDirWatch::INotify:
         entryAdded = useINotify(e);
@@ -978,35 +911,26 @@ void KDirWatchPrivate::addWatch(Entry *e)
         break;
     }
 
-    // Failing that try in order INotify, FAM, QFSWatch, Stat
+    // Failing that try in order INotify, QFSWatch, Stat
     if (!entryAdded) {
 #if HAVE_SYS_INOTIFY_H
-        if (useINotify(e)) {
-            return;
-        }
-#endif
-#if HAVE_FAM
-        if (useFAM(e)) {
+        if (preferredMethod != KDirWatch::INotify && useINotify(e)) {
             return;
         }
 #endif
 #if HAVE_QFILESYSTEMWATCHER
-        if (useQFSWatch(e)) {
+        if (preferredMethod != KDirWatch::QFSWatch && useQFSWatch(e)) {
             return;
         }
 #endif
-        useStat(e);
+        if (preferredMethod != KDirWatch::Stat) {
+            useStat(e);
+        }
     }
 }
 
 void KDirWatchPrivate::removeWatch(Entry *e)
 {
-#if HAVE_FAM
-    if (e->m_mode == FAMMode) {
-        FAMCancelMonitor(&fc, &(e->fr));
-        qCDebug(KDIRWATCH).nospace() << "Cancelled FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr)) << ") for " << e->path;
-    }
-#endif
 #if HAVE_SYS_INOTIFY_H
     if (e->m_mode == INotifyMode) {
         m_inotify_wd_to_entry.remove(e->wd);
@@ -1028,9 +952,8 @@ void KDirWatchPrivate::removeWatch(Entry *e)
 
 void KDirWatchPrivate::removeEntry(KDirWatch *instance, const QString &_path, Entry *sub_entry)
 {
-    if (s_verboseDebug) {
-        qCDebug(KDIRWATCH) << "path=" << _path << "sub_entry:" << sub_entry;
-    }
+    qCDebug(KDIRWATCH) << "path=" << _path << "sub_entry:" << sub_entry;
+
     Entry *e = entry(_path);
     if (e) {
         removeEntry(instance, e, sub_entry);
@@ -1071,7 +994,7 @@ void KDirWatchPrivate::removeEntry(KDirWatch *instance, Entry *e, Entry *sub_ent
     if (e->m_mode == StatMode) {
         statEntries--;
         if (statEntries == 0) {
-            timer.stop(); // stop timer if lists are empty
+            m_statRescanTimer.stop(); // stop timer if lists are empty
             qCDebug(KDIRWATCH) << " Stopped Polling Timer";
         }
     }
@@ -1114,8 +1037,8 @@ void KDirWatchPrivate::removeEntries(KDirWatch *instance)
     if (minfreq > freq) {
         // we can decrease the global polling frequency
         freq = minfreq;
-        if (timer.isActive()) {
-            timer.start(freq);
+        if (m_statRescanTimer.isActive()) {
+            m_statRescanTimer.start(freq);
         }
         qCDebug(KDIRWATCH) << "Poll Freq now" << freq << "msec";
     }
@@ -1249,7 +1172,7 @@ int KDirWatchPrivate::scanEntry(Entry *e)
         return NoChange;
     }
 
-    if (e->m_mode == FAMMode || e->m_mode == INotifyMode) {
+    if (e->m_mode == INotifyMode) {
         // we know nothing has changed, no need to stat
         if (!e->dirty) {
             return NoChange;
@@ -1417,7 +1340,7 @@ void KDirWatchPrivate::slotRemoveDelayed()
 {
     delayRemove = false;
     // Removing an entry could also take care of removing its parent
-    // (e.g. in FAM or inotify mode), which would remove other entries in removeList,
+    // (e.g. in inotify mode), which would remove other entries in removeList,
     // so don't use Q_FOREACH or iterators here...
     while (!removeList.isEmpty()) {
         Entry *entry = *removeList.begin();
@@ -1426,7 +1349,7 @@ void KDirWatchPrivate::slotRemoveDelayed()
 }
 
 /* Scan all entries to be watched for changes. This is done regularly
- * when polling. FAM and inotify use a single-shot timer to call this slot delayed.
+ * when polling. inotify uses a single-shot timer to call this slot delayed.
  */
 void KDirWatchPrivate::slotRescan()
 {
@@ -1440,9 +1363,9 @@ void KDirWatchPrivate::slotRescan()
     // like showing a message box. We don't want to keep polling during
     // that time, otherwise the value of 'delayRemove' will be reset.
     // ### TODO: now the emitEvent delays emission, this can be cleaned up
-    bool timerRunning = timer.isActive();
+    bool timerRunning = m_statRescanTimer.isActive();
     if (timerRunning) {
-        timer.stop();
+        m_statRescanTimer.stop();
     }
 
     // We delay deletions of entries this way.
@@ -1503,7 +1426,6 @@ void KDirWatchPrivate::slotRescan()
             }
             break;
 #endif
-        case FAMMode:
         case QFSWatchMode:
             if (ev == Created) {
                 addWatch(entry);
@@ -1538,7 +1460,7 @@ void KDirWatchPrivate::slotRescan()
     }
 
     if (timerRunning) {
-        timer.start(freq);
+        m_statRescanTimer.start(freq);
     }
 
 #if HAVE_SYS_INOTIFY_H
@@ -1572,216 +1494,57 @@ bool KDirWatchPrivate::isNoisyFile(const char *filename)
     return false;
 }
 
-void KDirWatchPrivate::ref()
+void KDirWatchPrivate::ref(KDirWatch *watch)
 {
-    ++m_references;
+    m_referencesObjects.push_back(watch);
 }
 
-void KDirWatchPrivate::unref()
+void KDirWatchPrivate::unref(KDirWatch *watch)
 {
-    --m_references;
-    if (m_references == 0) {
+    m_referencesObjects.removeOne(watch);
+    if (m_referencesObjects.isEmpty()) {
         destroyPrivate();
     }
 }
 
-#if HAVE_FAM
-void KDirWatchPrivate::famEventReceived()
+#if HAVE_SYS_INOTIFY_H
+QString KDirWatchPrivate::inotifyEventName(const inotify_event *event) const
 {
-    static FAMEvent fe;
-
-    delayRemove = true;
-
-    while (use_fam && FAMPending(&fc)) {
-        if (FAMNextEvent(&fc, &fe) == -1) {
-            disableFAM();
-        } else {
-            checkFAMEvent(&fe);
-        }
-    }
-
-    QTimer::singleShot(0, this, &KDirWatchPrivate::slotRemoveDelayed);
-}
-
-void KDirWatchPrivate::disableFAM()
-{
-    qCWarning(KCOREADDONS_DEBUG) << "FAM connection problem, switching to a different system.";
-    use_fam = false;
-    delete sn;
-    sn = nullptr;
-
-    // Replace all FAMMode entries with another system (INotify/QFSW/Stat)
-    for (auto it = m_mapEntries.begin(); it != m_mapEntries.end(); ++it) {
-        if ((*it).m_mode == FAMMode && !(*it).m_clients.empty()) {
-            Entry *e = &(*it);
-            addWatch(e);
-        }
-    }
-}
-
-void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
-{
-    Entry *e = nullptr;
-    EntryMap::Iterator it = m_mapEntries.begin();
-    for (; it != m_mapEntries.end(); ++it)
-        if (FAMREQUEST_GETREQNUM(&((*it).fr)) == FAMREQUEST_GETREQNUM(&(fe->fr))) {
-            e = &(*it);
-            break;
-        }
-
-    // Don't be too verbose ;-)
-    if ((fe->code == FAMExists) || (fe->code == FAMEndExist) || (fe->code == FAMAcknowledge)) {
-        if (e) {
-            e->m_famReportedSeen = true;
-        }
-        return;
-    }
-
-    if (isNoisyFile(fe->filename)) {
-        return;
-    }
-
-    // Entry *e = static_cast<Entry*>(fe->userdata);
-
-    if (s_verboseDebug) { // don't enable this except when debugging, see #88538
-        qCDebug(KDIRWATCH) << "Processing FAM event ("
-                           << ((fe->code == FAMChanged)              ? "FAMChanged"
-                                   : (fe->code == FAMDeleted)        ? "FAMDeleted"
-                                   : (fe->code == FAMStartExecuting) ? "FAMStartExecuting"
-                                   : (fe->code == FAMStopExecuting)  ? "FAMStopExecuting"
-                                   : (fe->code == FAMCreated)        ? "FAMCreated"
-                                   : (fe->code == FAMMoved)          ? "FAMMoved"
-                                   : (fe->code == FAMAcknowledge)    ? "FAMAcknowledge"
-                                   : (fe->code == FAMExists)         ? "FAMExists"
-                                   : (fe->code == FAMEndExist)       ? "FAMEndExist"
-                                                                     : "Unknown Code")
-                           << ", " << fe->filename << ", Req " << FAMREQUEST_GETREQNUM(&(fe->fr)) << ") e=" << e;
-    }
-
-    if (!e) {
-        // this happens e.g. for FAMAcknowledge after deleting a dir...
-        //    qCDebug(KDIRWATCH) << "No entry for FAM event ?!";
-        return;
-    }
-
-    if (e->m_status == NonExistent) {
-        qCDebug(KDIRWATCH) << "FAM event for nonExistent entry " << e->path;
-        return;
-    }
-
-    // Delayed handling. This rechecks changes with own stat calls.
-    e->dirty = true;
-    if (!rescan_timer.isActive()) {
-        rescan_timer.start(m_PollInterval); // singleshot
-    }
-
-    // needed FAM control actions on FAM events
-    switch (fe->code) {
-    case FAMDeleted:
-        // fe->filename is an absolute path when a watched file-or-dir is deleted
-        if (!QDir::isRelativePath(QFile::decodeName(fe->filename))) {
-            FAMCancelMonitor(&fc, &(e->fr)); // needed ?
-            qCDebug(KDIRWATCH) << "Cancelled FAMReq" << FAMREQUEST_GETREQNUM(&(e->fr)) << "for" << e->path;
-            e->m_status = NonExistent;
-            e->m_ctime = invalid_ctime;
-            emitEvent(e, Deleted, e->path);
-            // If the parent dir was already watched, tell it something changed
-            Entry *parentEntry = entry(e->parentDirectory());
-            if (parentEntry) {
-                parentEntry->dirty = true;
-            }
-            // Add entry to parent dir to notice if the entry gets recreated
-            addEntry(nullptr, e->parentDirectory(), e, true /*isDir*/);
-        } else {
-            // A file in this directory has been removed, and wasn't explicitly watched.
-            // We could still inform clients, like inotify does? But stat can't.
-            // For now we just marked e dirty and slotRescan will emit the dir as dirty.
-            // qCDebug(KDIRWATCH) << "Got FAMDeleted for" << QFile::decodeName(fe->filename) << "in" << e->path << ". Absolute path -> NOOP!";
-        }
-        break;
-
-    case FAMCreated: {
-        // check for creation of a directory we have to watch
-        QString tpath(e->path + QLatin1Char('/') + QFile::decodeName(fe->filename));
-
-        // This code is very similar to the one in inotifyEventReceived...
-        Entry *sub_entry = e->findSubEntry(tpath);
-        if (sub_entry /*&& sub_entry->isDir*/) {
-            // We were waiting for this new file/dir to be created.  We don't actually
-            // emit an event here, as the rescan_timer will re-detect the creation and
-            // do the signal emission there.
-            sub_entry->dirty = true;
-            rescan_timer.start(0); // process this asap, to start watching that dir
-        } else if (e->isDir && !e->m_clients.empty()) {
-            bool isDir = false;
-            const QList<const Client *> clients = e->clientsForFileOrDir(tpath, &isDir);
-            for (const Client *client : clients) {
-                addEntry(client->instance, tpath, nullptr, isDir, isDir ? client->m_watchModes : KDirWatch::WatchDirOnly);
-            }
-
-            if (!clients.isEmpty()) {
-                emitEvent(e, Created, tpath);
-
-                qCDebug(KDIRWATCH).nospace() << clients.count() << " instance(s) monitoring the new " << (isDir ? "dir " : "file ") << tpath;
-            }
-        }
-    } break;
-    default:
-        break;
-    }
-}
-#else
-void KDirWatchPrivate::famEventReceived()
-{
-    qCWarning(KCOREADDONS_DEBUG) << "Fam event received but FAM is not supported";
+    if (event->mask & IN_OPEN)
+        return QStringLiteral("OPEN");
+    else if (event->mask & IN_CLOSE_NOWRITE)
+        return QStringLiteral("CLOSE_NOWRITE");
+    else if (event->mask & IN_CLOSE_WRITE)
+        return QStringLiteral("CLOSE_WRITE");
+    else if (event->mask & IN_MOVED_TO)
+        return QStringLiteral("MOVED_TO");
+    else if (event->mask & IN_MOVED_FROM)
+        return QStringLiteral("MOVED_FROM");
+    else if (event->mask & IN_MOVE)
+        return QStringLiteral("MOVE");
+    else if (event->mask & IN_CREATE)
+        return QStringLiteral("CREATE");
+    else if (event->mask & IN_DELETE)
+        return QStringLiteral("DELETE");
+    else if (event->mask & IN_DELETE_SELF)
+        return QStringLiteral("DELETE_SELF");
+    else if (event->mask & IN_MOVE_SELF)
+        return QStringLiteral("MOVE_SELF");
+    else if (event->mask & IN_ATTRIB)
+        return QStringLiteral("ATTRIB");
+    else if (event->mask & IN_MODIFY)
+        return QStringLiteral("MODIFY");
+    if (event->mask & IN_ACCESS)
+        return QStringLiteral("ACCESS");
+    if (event->mask & IN_IGNORED)
+        return QStringLiteral("IGNORED");
+    if (event->mask & IN_UNMOUNT)
+        return QStringLiteral("IN_UNMOUNT");
+    else
+        return QStringLiteral("UNKWOWN");
 }
 #endif
 
-void KDirWatchPrivate::statistics()
-{
-    EntryMap::Iterator it;
-
-    qCDebug(KDIRWATCH) << "Entries watched:";
-    if (m_mapEntries.count() == 0) {
-        qCDebug(KDIRWATCH) << "  None.";
-    } else {
-        it = m_mapEntries.begin();
-        for (; it != m_mapEntries.end(); ++it) {
-            Entry *e = &(*it);
-            qCDebug(KDIRWATCH) << "  " << *e;
-
-            for (const Client &c : e->m_clients) {
-                QByteArray pending;
-                if (c.watchingStopped) {
-                    if (c.pending & Deleted) {
-                        pending += "deleted ";
-                    }
-                    if (c.pending & Created) {
-                        pending += "created ";
-                    }
-                    if (c.pending & Changed) {
-                        pending += "changed ";
-                    }
-                    if (!pending.isEmpty()) {
-                        pending = " (pending: " + pending + ')';
-                    }
-                    pending = ", stopped" + pending;
-                }
-                qCDebug(KDIRWATCH) << "    by " << c.instance->objectName() << " (" << c.count << " times)" << pending;
-            }
-            if (!e->m_entries.isEmpty()) {
-                qCDebug(KDIRWATCH) << "    dependent entries:";
-                for (Entry *d : std::as_const(e->m_entries)) {
-                    qCDebug(KDIRWATCH) << "      " << d << d->path << (d->m_status == NonExistent ? "NonExistent" : "EXISTS!!! ERROR!");
-                    if (s_verboseDebug) {
-                        Q_ASSERT(d->m_status == NonExistent); // it doesn't belong here otherwise
-                    }
-                }
-            }
-        }
-    }
-}
-
 #if HAVE_QFILESYSTEMWATCHER
 // Slot for QFileSystemWatcher
 void KDirWatchPrivate::fswEventReceived(const QString &path)
@@ -1851,33 +1614,21 @@ bool KDirWatch::exists()
     return s_pKDirWatchSelf.exists() && dwp_self.hasLocalData();
 }
 
-static void postRoutine_KDirWatch()
-{
-    if (s_pKDirWatchSelf.exists()) {
-        s_pKDirWatchSelf()->deleteQFSWatcher();
-    }
-}
-
 KDirWatch::KDirWatch(QObject *parent)
     : QObject(parent)
     , d(createPrivate())
 {
-    d->ref();
+    d->ref(this);
     static QBasicAtomicInt nameCounter = Q_BASIC_ATOMIC_INITIALIZER(1);
     const int counter = nameCounter.fetchAndAddRelaxed(1); // returns the old value
     setObjectName(QStringLiteral("KDirWatch-%1").arg(counter));
-
-    if (counter == 1) { // very first KDirWatch instance
-        // Must delete QFileSystemWatcher before qApp is gone - bug 261541
-        qAddPostRoutine(postRoutine_KDirWatch);
-    }
 }
 
 KDirWatch::~KDirWatch()
 {
-    if (d && dwp_self.hasLocalData()) { // skip this after app destruction
+    if (d) {
         d->removeEntries(this);
-        d->unref();
+        d->unref(this);
     }
 }
 
@@ -1991,22 +1742,6 @@ bool KDirWatch::contains(const QString &_path) const
     return false;
 }
 
-void KDirWatch::deleteQFSWatcher()
-{
-    delete d->fsWatcher;
-    d->fsWatcher = nullptr;
-    d = nullptr;
-}
-
-void KDirWatch::statistics()
-{
-    if (!dwp_self.hasLocalData()) {
-        qCDebug(KDIRWATCH) << "KDirWatch not used";
-        return;
-    }
-    dwp_self.localData()->statistics();
-}
-
 void KDirWatch::setCreated(const QString &_file)
 {
     qCDebug(KDIRWATCH) << objectName() << "emitting created" << _file;
@@ -2028,13 +1763,6 @@ KDirWatch::Method KDirWatch::internalMethod() const
 {
     // This reproduces the logic in KDirWatchPrivate::addWatch
     switch (d->m_preferredMethod) {
-    case KDirWatch::FAM:
-#if HAVE_FAM
-        if (d->use_fam) {
-            return KDirWatch::FAM;
-        }
-#endif
-        break;
     case KDirWatch::INotify:
 #if HAVE_SYS_INOTIFY_H
         if (d->supports_inotify) {
@@ -2057,11 +1785,6 @@ KDirWatch::Method KDirWatch::internalMethod() const
         return KDirWatch::INotify;
     }
 #endif
-#if HAVE_FAM
-    if (d->use_fam) {
-        return KDirWatch::FAM;
-    }
-#endif
 #if HAVE_QFILESYSTEMWATCHER
     return KDirWatch::QFSWatch;
 #else
@@ -2069,6 +1792,36 @@ KDirWatch::Method KDirWatch::internalMethod() const
 #endif
 }
 
+bool KDirWatch::event(QEvent *event)
+{
+    if (Q_LIKELY(event->type() != QEvent::ThreadChange)) {
+        return QObject::event(event);
+    }
+
+    qCCritical(KDIRWATCH) << "KDirwatch is moving its thread. This is not supported at this time; your watch will not watch anything anymore!"
+                          << "Create and use watches on the correct thread"
+                          << "Watch:" << this;
+
+    // We are still in the old thread when the event runs, so this is safe.
+    Q_ASSERT(thread() == d->thread());
+    d->removeEntries(this);
+    d->unref(this);
+    d = nullptr;
+
+    // Schedule the creation of the new private in the new thread.
+    QMetaObject::invokeMethod(
+        this,
+        [this] {
+            d = createPrivate();
+        },
+        Qt::QueuedConnection);
+
+    // NOTE: to actually support moving watches across threads we'd have to make Entry copyable and schedule a complete
+    // re-installation of watches on the new thread after createPrivate.
+
+    return QObject::event(event);
+}
+
 #include "moc_kdirwatch.cpp"
 #include "moc_kdirwatch_p.cpp"
 
