@@ -268,11 +268,7 @@ QString KAboutLicense::text() const
     }
 
     if (knownLicense) {
-#ifndef Q_OS_ANDROID
-        pathToFile = QStandardPaths::locate(QStandardPaths::GenericDataLocation, QStringLiteral("kf6/licenses/") + pathToFile);
-#else
-        pathToFile = QStringLiteral("assets:/share/kf6/licenses/") + pathToFile;
-#endif
+        pathToFile = QStringLiteral(":/org.kde.kcoreaddons/licenses/") + pathToFile;
         result += QCoreApplication::translate("KAboutLicense", "This program is distributed under the terms of the %1.").arg(name(KAboutLicense::ShortName));
         if (!pathToFile.isEmpty()) {
             result += lineFeed;
